

var mainPanel, aboutPanel, sharePanel,
iterEle, testTemplate, testsEle, runTestBtn, testResultsEle, shareURL,
tests = [], syntaxHL, runningTest = -1, testFrames = 0,
loadingSharedURL = false, exportDataURL = null, compressor,
vertexShader, testStartTime = 0, testDoneFence,

COMPRESSION_PREMEMORY = "for while length(atan(sqrt(dot(sin(cos( return; void int uint float vec2 vec3 vec4 mat2 mat3 mat4 ^=%=&=|=+=-=*=/= < > & | ^ / * - + = \n",
SHADER_HEADER = "#version 300 es\nprecision highp float;precision highp int;\n";


//DOM loaded, program entry point
document.addEventListener("DOMContentLoaded",function() {	

	//load webgl 2
	if (!InitializeQuickWebGL(document.createElement("canvas"))) {
		FatalCrash("Fatal Error: Your browser does not support WebGL 2. Try again on a modern browser such as Firefox or Chrome and with a device capable of at least OpenGL ES 3.0 or Core 3.2.");
		return;
	}
	vertexShader = CompileShader(SHADER_HEADER+"in vec2 vpos;out vec2 uv;void main() {uv = vpos*.5+.5;gl_Position = vec4(vpos,0,1);}", gl.VERTEX_SHADER);
	BindMesh(fullscreenTriangle);

	//detect if loading shared url
	if (location.hash && location.hash.length > 2) {
		loadingSharedURL = true;
	
	} else {
		//load last tests from storage or add default
		IndexedStorage_Open("BenchmarkGLSL", function() {
			IndexedStorage_Load("Last", function(id,data) {
				if (data) {
					DeserializeTests(data);
				
				} else {
					var t = new Test();
					tests.push(t);
					t.nameEle.value = "sqrt()";
					t.editor.SetText(
'float Exec(vec3 seed) {//execute test\n'+
	'\tfloat x = length(seed);\n'+
	'\treturn sqrt(x);\n'+
'}');
					
					t = new Test();
					tests.push(t);
					t.nameEle.value = "Babylonian Sqrt";
					t.editor.SetText(
'float Exec(vec3 seed) {//execute test\n'+
	'\tfloat x = length(seed), s = x-1e-6;\n'+
	'\tfor (int i = 0; i < 50; i++) s = (s+x/s)*.5;\n'+
	'\treturn s;\n'+
'}');

				}
			});
		});
	}
	

	//glsl syntax highlighting
	InitializeSyntaxHighlighting();
	syntaxHL = new SyntaxHighlighting("hl-c","hl-s","hl-q","hl-a","hl-p","hl-d",LANGUAGE_GLSL);
	
	//get elements
	mainPanel = document.getElementById("MainPanel");
	aboutPanel = document.getElementById("AboutPanel");
	sharePanel = document.getElementById("SharePanel");
	
	testsEle = document.getElementById("Tests");
	testResultsEle = document.getElementById("TestResults");
	iterEle = document.getElementById("Iter");
	shareURL = document.getElementById("ShareURL");

	//test gui template
	testTemplate = new GUITemplate("TestTemplate");
	
	
	//add new test button event
	BEC("Add", function() {
		if (tests.length >= 255) {
			testResultsEle.innerHTML = "Error failed to create test, reached maximum of 255 tests.";
			return;
		}
		tests.push(new Test());
	});
	
	
	//run tests
	BEC("Run", function(e) {
		runTestBtn = e.target;
		
		if (runningTest === -1) {
			//run tests
			if (tests.length === 0) return;
			
			//save tests to storage
			if (!loadingSharedURL) IndexedStorage_Store("Last", SerializeTests());
			
			//compile test shaders
			for (var i = 0; i < tests.length; i++) {
				var t = tests[i],
					s = gl.createShader(gl.FRAGMENT_SHADER);
				gl.shaderSource(s, SHADER_HEADER+t.editor.text.value+"in vec2 uv;uniform float frame;out vec4 fout;\nvoid main() {fout = vec4(Exec(vec3(uv,frame)));}");
				gl.compileShader(s);
				if (!gl.getShaderParameter(s,gl.COMPILE_STATUS)) {
					//compilation error, display errors
					testResultsEle.innerHTML = "Error running test #"+(i+1)+" '"+t.nameEle.value+".</br><span style='color:#ff8020;'>"+gl.getShaderInfoLog(s)+"</span>";

					gl.deleteShader(s);
					for (var k = 0; k < i; k++) gl.deleteProgram(tests[k].shader);
					return;
				}
				
				var p = t.shader = gl.createProgram();
				gl.attachShader(p, vertexShader);
				gl.attachShader(p, s);
				gl.bindAttribLocation(p,0,"vpos");
				gl.linkProgram(p);
				gl.detachShader(p, vertexShader);
				gl.detachShader(p, s);
				gl.deleteShader(s);
				if (!gl.getProgramParameter(p,gl.LINK_STATUS)) {
					//linking error, display errors
					testResultsEle.innerHTML = "Error running test #"+(i+1)+" '"+t.nameEle.value+".</br><span style='color:#ff8020;'>"+gl.getProgramInfoLog(p)+"</span>";

					gl.deleteProgram(p);
					for (var k = 0; k < i; k++) gl.deleteProgram(tests[k].shader);
					return;
				}
				t.frameUni = gl.getUniformLocation(p, "frame");
			}
			
			//split iterations into size*size*frames, prioritizing size to reduce draw call overhead
			var iter = ReadIterCount(),
				sz = Math.min(2048, Math.max(2, Math.round(Math.sqrt(iter))));
			quickGLCanvas.width = quickGLCanvas.height = display.width = display.height = sz;
			testFrames = Math.ceil(iter/(sz*sz));
			
			//launch
			display.Bind();
			runningTest = 0;
			LaunchTest();
			
		} else {
			//stop tests
			testResultsEle.innerHTML = "";
			runningTest = -1;
		}
		UpdateRunTestBtn();
	});
	
	
	//share
	BEC("Share", function() {
		EV(mainPanel,false);
		EV(sharePanel,true);
		
		//generate share url
		shareURL.textContent = "";
		setTimeout(GenerateShare,0);
	});
	
	//copy to clipboard
	BEC("ShareCopy", function() {
		navigator.clipboard.writeText(shareURL.textContent);
	});
	
	//close share
	BEC("ShareBack", function() {
		EV(sharePanel,false);
		EV(mainPanel,true);
	});
	
	
	//export
	BEC("Export", function() {
		if (exportDataURL) URL.revokeObjectURL(exportDataURL);
		exportDataURL = URL.createObjectURL(new Blob([SerializeTests()]));
		SaveFilePrompt(exportDataURL, "Test.benchglsl");
	});
	
	//import
	BEC("Import", function() {
		LoadFilePrompt(OnImportFileSelect, ".benchglsl");
	});
	
	
	//open about page
	BEC("About", function(e) {
		EV(mainPanel,false);
		EV(aboutPanel,true);
	});
	
	//close about page
	BEC("AboutBack", function(e) {
		EV(aboutPanel,false);
		EV(mainPanel,true);
	});
	
	
	
	//load shared url data
	if (loadingSharedURL) DeserializeTests(DecodeURLBase64(location.hash.substring(1)), true);
});


//display full screen crash message
function FatalCrash(msg) {
	document.body.innerHTML = "<div class='FatalCrash'>"+msg+"</div>";
}


//Test class
/**@constructor*/
function Test() {
	var tele = testTemplate.Clone();
	this.element = tele;
	this.nameEle = GUITemplate_Get(tele, testTemplate["TestName"]);
	this.editor = new CodeEditor(GUITemplate_Get(tele, testTemplate["TestCode"]), "\t",true,true, GUITemplate_Get(tele, testTemplate["TestCodeHL"]),syntaxHL,0);
	this.editor.UpdateSyntaxHighlighting();
	this.eps = 0;
	this.nameLen = 0;
	this.codeLen = 0;
	this.shader = null;
	this.frameUni = null;
	
	var delBtn = GUITemplate_Get(tele, testTemplate["TestDelete"]);
	delBtn.testObjAttach = this;
	delBtn.addEventListener("click", OnDeleteTest);
	
	testsEle.appendChild(tele);
}

//delete test
function OnDeleteTest(e) {
	var test = e.target.testObjAttach,
		tid = tests.indexOf(test);
	testsEle.removeChild(test.element);
	tests.splice(tid,1);
}

//update run tests btn text
function UpdateRunTestBtn() {
	runTestBtn.textContent = (runningTest===-1 ? "Run Tests" : "Stop");
}

//cmp function for test sorting
function TestSortCmp(a,b) {
	return tests[a].eps < tests[b].eps;
}

//read iteration count element
function ReadIterCount() {
	var iv = parseFloat(iterEle.value);
	return isNaN(iv) ? 1 : Math.max(1,Math.floor(iv));
}


//launch next test
function LaunchTest() {	
	var test = tests[runningTest];
	testResultsEle.innerHTML = "Running test #"+(runningTest+1)+" '"+test.nameEle.value+"'...";

	setTimeout(RunTest,0);
}

//execute test
function RunTest() {
	var test = tests[runningTest],
		tuni = test.frameUni;
	gl.useProgram(test.shader);
	
	testStartTime = Date.now();
	for (var i = 0; i < testFrames; i++) {
		gl.uniform1f(tuni, i/testFrames);
		gl.drawArrays(gl.TRIANGLES,0,3);
	}
	testDoneFence = gl.fenceSync(gl.SYNC_GPU_COMMANDS_COMPLETE, 0);

	setTimeout(WaitTest,0);
}

//wait for test complete
function WaitTest() {
	if (runningTest === -1) {
		//stopped
		gl.deleteSync(testDoneFence);
		return;
	}
	
	var res = gl.clientWaitSync(testDoneFence,gl.SYNC_FLUSH_COMMANDS_BIT,0);
	if (res !== gl.TIMEOUT_EXPIRED) {
		//done
		var endTime = Date.now();
		gl.deleteSync(testDoneFence);
		
		tests[runningTest].eps = (display.width*display.height*testFrames)/Math.max(1e-6, (endTime-testStartTime)/1000);

		if (++runningTest >= tests.length) {
			//finished
			runningTest = -1;
			UpdateRunTestBtn();
		
			//sort tests best to worst and delete test shaders
			var order = new Array(tests.length);
			for (var i = 0; i < tests.length; i++) {
				order[i] = i;
				gl.deleteProgram(tests[i].shader);
			}
			order.sort(TestSortCmp);
		
			//display results in table
			var h = "<table><tr><th>Test</th> <th>Executions Per Second</th> <th>Scale</th></tr>",
				worstEps = Math.max(1e-8,tests[order[tests.length-1]].eps);
			for (var i = 0; i < tests.length; i++) {
				var t = tests[order[i]];
				h += "<tr><td>"+t.nameEle.value+"</td><td>"+t.eps.toLocaleString()+"</td><td>"+(t.eps/worstEps).toLocaleString()+"x</td></tr>";
			}
			h += "</table>";
			testResultsEle.innerHTML = h;
			return;
		}
		LaunchTest();
		return;
	}
	
	setTimeout(WaitTest,0);
}


//serialize tests to binary string
function SerializeTests(compress) {
	var o = String.fromCharCode(tests.length);//num tests as byte
	
	o += VarLenIntWrite(ReadIterCount());//encode iterations as variable length integer 
	
	//combine all test names/code into one big string and write out lengths(to calculate indices)
	var sum = "";
	for (var i = 0; i < tests.length; i++) {
		var test = tests[i],
			tname = test.nameEle.value,
			tcode = test.editor.text.value;
		o += VarLenIntWrite(tname.length);
		o += VarLenIntWrite(tcode.length);
		sum += tname+tcode;
	}
	
	//compress
	if (compress) {
		if (!compressor) compressor = new Compression();
		var cmp = compressor.Compress(COMPRESSION_PREMEMORY+sum, COMPRESSION_MAX_MEMORY, COMPRESSION_PREMEMORY.length);
		sum = cmp;
	}
	
	return o+sum;
}

//deserialize tests from binary string
function DeserializeTests(o, compress) {
	var ntests = o.charCodeAt(0);
	
	iterEle.value = VarLenIntRead(o,1);
	
	//delete old tests
	for (var i = 0; i < tests.length; i++) testsEle.removeChild(tests[i].element);
	tests.length = 0;
	
	//allocate tests and read lengths
	for (var i = 0; i < ntests; i++) {
		var t = new Test();
		t.nameLen = VarLenIntRead(o,varLenIntIndex);
		t.codeLen = VarLenIntRead(o,varLenIntIndex);
		tests.push(t);
	}
	
	//decompress string data
	var sum = o.substring(varLenIntIndex);
	if (compress) {
		if (!compressor) compressor = new Compression();
		sum = compressor.Decompress(sum,COMPRESSION_PREMEMORY);
	}
	
	//extract strings
	var ind = 0;
	for (var i = 0; i < ntests; i++) {
		var t = tests[i], nind = ind+t.nameLen;
		t.nameEle.value = sum.substring(ind,nind);
		ind = nind;
		nind += t.codeLen;
		t.editor.SetText(sum.substring(ind,nind));
		ind = nind;
	}
}


//convert between binary string and uint8array
function StringToUint8(s) {
	var sl = s.length, u = new Uint8Array(sl);
	for (var i = 0; i < sl; i++) u[i] = s.charCodeAt(i);
	return u;
}
function Uint8ToString(a) {
	var s = "", sl = a.length;
	for (var i = 0; i < sl; i++) s += String.fromCharCode(a[i]);
	return s;
}


if (!String.prototype.replaceAll) {
	String.prototype.replaceAll = function(find,replac) {
		return this.replace(new RegExp(find,"g"), replac);
	}
}

//encode binary string as url safe base64, returns base64
function EncodeURLBase64(data) {
	var b = btoa(data), end = b.length-1, bi = end;
	//remove padding
	while (b[bi] === "=") bi--;
	if (bi < end) b = b.substring(0,bi+1); 
	//replace unsafe chars
	return b.replaceAll("+","-").replaceAll("/","_");
}

//decode url safe base64 into binary string, returns binary string
function DecodeURLBase64(data) {
	//replace unsafe chars
	data = data.replaceAll("-","+").replaceAll("_","/");
	//add padding back
	var pad = data.length%4;
	if (pad) {
		pad = 4-pad;
		for (var i = 0; i < pad; i++) data += "=";
	}
	return atob(data);//decode
}



//on import file select
function OnImportFileSelect(e) {
	var f = e.target.files;
	if (f && f.length) {
		var fread = new FileReader();
		fread.onload = OnImportFileLoad;
		fread.readAsText(f[0]);
	}
}

//on import file load
function OnImportFileLoad(e) {
	DeserializeTests(e.target.result);
}


//generate shareable url
function GenerateShare() {
	shareURL.textContent = "https://xaloez.com/o/BenchmarkGLSL#"+EncodeURLBase64(SerializeTests(true));
}
