
var COMPRESSION_MAX_MEMORY = 8229;


//Compression class
/**@constructor*/
function Compression() {
	this.last = new Array(256);
	this.last.fill(-1);
	this.lut = new Array(256);
	for (var i = 0; i < 256; i++) this.lut[i] = new Array(256);

	this.output = null;
	this.input = null;
	this.inputIndex = 0;
	this.byte = 0;
	this.bit = 0;
}


//reading and writing serialization functions used by compression class, not meant to be used outside
//output bit b
Compression.prototype.WB = function(b) {
	this.byte |= (b<<this.bit);
	if (++this.bit >= 8) {
		this.output += String.fromCharCode(this.byte);
		this.bit = this.byte = 0;
	}
}

//output c bits from integer n
Compression.prototype.WN = function(n,c) {
	while (c) {
		var oc = Math.min(8-this.bit, c);
		
		this.byte |= (n&((1<<oc)-1))<<this.bit;
		this.bit += oc;
		if (this.bit >= 8) {
			this.output += String.fromCharCode(this.byte);
			this.bit = this.byte = 0;
		}
		
		n >>= oc;
		c -= oc;

	}
}

//output integer n with variable length encoding, n must be below 8229
Compression.prototype.WV = function(n) {
	if (n === 0) {
		this.WB(0);
		
	} else if (n < 5) {//= (1<<2)+1
		this.WN(((n-1)<<2) | 1, 4);
	
	} else if (n < 37) {//= (1<<5)+5
		this.WN(((n-5)<<3) | 3, 8);
		
	} else {//if (n < 8229) {//= (1<<13)+37
		this.WN(((n-37)<<3) | 7, 16);
	}
}
function Compression_VLSize(n) {
	if (n === 0) return 1;
	else if (n < 5) return 4;
	else if (n < 37) return 8;
	else return 16;
}


//read bit from input
Compression.prototype.RB = function() {
	var b = (this.byte>>this.bit)&1;
	if (++this.bit >= 8) {
		this.bit = 0;
		
		var inp = this.input;
		if (++this.inputIndex < inp.length) {
			this.byte = inp.charCodeAt(this.inputIndex);
		} else {
			this.byte = -1;
		}
	}
	return b;
}

//read c bit integer from input
Compression.prototype.RN = function(c) {
	var n = 0, end = c;
	while (c) {
		var oc = Math.min(8-this.bit, c);
		n |= ((this.byte>>this.bit)&((1<<oc)-1))<<(end-c);
		
		this.bit += oc;
		if (this.bit >= 8) {
			this.bit = 0;
			
			var inp = this.input;
			if (++this.inputIndex < inp.length) {
				this.byte = inp.charCodeAt(this.inputIndex);
			} else {
				this.byte = -1;
				return n;
			}
		}

		c -= oc;
	}
	return n;
}

//read variable length integer from input
Compression.prototype.RV = function() {
	if (this.RB()) {
		if (this.RB()) {
			if (this.RB()) return this.RN(13)+37;
			else return this.RN(5)+5;
			
		} else return this.RN(2)+1;
	
	} else return 0;
}



/*Compress binary string input, returns compressed binary string.
memorySize is how many bytes back can be recalled for compression, must be between 3 and 8229(inclusive).
preMemorySize is a number of bytes at the start of input that is used to optimize compression by providing some preset sequences, it is not included in the compressed output.*/
Compression.prototype.Compress = function(input, memorySize, preMemorySize) {
	if (input.length === 0) return "";

	var ilen = input.length, ilen1 = ilen-1, i = 0,
		newChunk = false, newStart = 0, lastCh = -1,
		last = this.last, lut = this.lut;
	this.output = "";
	if (!preMemorySize) preMemorySize = 0;
	
	while (true) {
		//loop through characters
		var ch = input.charCodeAt(i), step = 1;
		
		if (i >= preMemorySize) {
			var lastInd = last[ch],
				bestSeq = 0, bestInd = 0, bestLen = 0;
			if (lastInd !== -1) {//if repeat of 1 of 4 last characters we can encode in 4 bits
				lastInd = i-lastInd-1;
				if (lastInd < 4) {
					bestSeq = 0.25;//= 1/4
					bestInd = lastInd;
					bestLen = 0;
				}
			}
			
			//look for upcoming sequence in past sequences and find which has best compression ratio
			if (i !== ilen1) {
				var chLut = lut[ch], seqLut = chLut[input.charCodeAt(i+1)];
				if (seqLut) {
					for (var s = 0; s < seqLut.length; s++) {
						var sind = seqLut[s], k, ii = i+2;
						for (k = sind; k < i; k++) {
							if (input[k] !== input[ii++]) break;
						}
				
						var slen = k-sind+1,//-1
							offset = i-k,
							outSz = Compression_VLSize(slen)+Compression_VLSize(offset)+1;
				
						var bytesPerBit = (slen+1)/outSz;
						if (bytesPerBit > bestSeq) {
							bestSeq = bytesPerBit;
							bestInd = offset;
							bestLen = slen;
						}
					}
				}
			}
		
			//choose chunk based on best compression rate
			if (bestSeq > 0.1) {//= 1/10(worst case new chunk rate)
				//close new data chunk
				if (newChunk) {
					var newSz = i-newStart;
					while (newSz > 0) {
						var nsz = Math.min(8229,newSz);
						this.WB(0);
						this.WV(nsz-1);
					
						var end = newStart+nsz;
						for (newStart; newStart < end; newStart++) this.WN(input.charCodeAt(newStart), 8);
						newSz -= nsz;
					}
					newChunk = false;
				}
		
				//repeat past sequence/char
				this.WB(1);
				this.WV(bestLen);
				if (bestLen === 0) this.WN(bestInd,2);
				else this.WV(bestInd);
			
				step = bestLen+1;
			
			} else {
				//new data
				if (!newChunk) {
					//start new data chunk
					newChunk = true;
					newStart = i;
				}
			}
		}
		
		//step char index forward
		var ni = i+step;
		if (ni === ilen) {
			i = ni;//done
			break;
		}
		
		while (true) {
			//remove previously recorded chars now outside memory range
			if (i >= memorySize) {
				var oldInd = i-memorySize,
					oldCh = input.charCodeAt(oldInd),
					seqLut = lut[input.charCodeAt(oldInd)][input.charCodeAt(oldInd+1)],
					seqInd = seqLut.indexOf(oldInd+2);
				if (seqInd !== -1) {
					var eind = seqLut.length-1;
					if (eind !== seqInd) seqLut[seqInd] = seqLut[eind];
					seqLut.length = eind;
				}
			}
			
			//record chars in lut
			last[ch] = i;
			if (lastCh !== -1) {
				var chLut = lut[lastCh],
					seqLut = chLut[ch];
				if (seqLut) seqLut.push(i+1);
				else chLut[ch] = [i+1];
			}
			
			lastCh = ch;
			i++;
			if (i >= ni) break;
			ch = input.charCodeAt(i);
		}
	}
	
	//close new data chunk
	if (newChunk) {
		var newSz = i-newStart;
		while (newSz > 0) {
			var nsz = Math.min(8229,newSz);
			this.WB(0);
			this.WV(nsz-1);
		
			var end = newStart+nsz;
			for (newStart; newStart < end; newStart++) this.WN(input.charCodeAt(newStart), 8);
			newSz -= nsz;
		}
	}
	
	//if not byte aligned fill with padding
	if (this.bit) {
		this.WB(0);//new chunk with 2 byte length which will never be reached
		while (this.bit) this.WB(1);
	}
	var out = this.output;
	
	//reset for next time
	this.output = null;
	this.bit = this.byte = 0;
	last.fill(-1);
	for (var i = 0; i < 256; i++) {
		var lt = lut[i];
		for (var k = 0; k < 256; k++) {
			var a = lt[k];
			if (a) a.length = 0;
		}
	}
	
	return out;
}



//Decompress input, returns decompressed output.
Compression.prototype.Decompress = function(input, preMemory) {
	this.output = (preMemory ? preMemory : "");
	this.input = input;
	this.inputIndex = 0;
	this.byte = input.charCodeAt(0);
	
	//decompress
	while (this.byte !== -1) {
		if (this.RB()) {//repeat data chunk
			var sz = this.RV()+1, off;
			if (sz === 1) off = 1+this.RN(2);
			else off = sz+this.RV();

			off = this.output.length-off;
			for (sz += off; off < sz; off++) this.output += this.output[off];
		
		} else {//new data chunk
			var sz = this.RV();
			if (this.byte === -1) break;
			sz++;
			for (var i = 0; i < sz; i++) this.output += String.fromCharCode(this.RN(8));
		}
	}
	
	var out = (preMemory ? this.output.substring(preMemory.length) : this.output);
	
	//reset for next time
	this.output = this.input = null;
	this.bit = this.byte = 0;
	
	return out;
}




//write integer with variable length from 1-9 bytes, returns binary string
function VarLenIntWrite(i) {
	var o = "";
	for (var mb = 0; mb < 9; mb++) {
		var fits = i<128;
		o += String.fromCharCode((fits?128:0) | (i&127));
		
		if (fits) break;
		i = Math.floor(i/128);
	}
	return o;
}

//read variable length integer from binary string, returns int and outputs new index in varLenIntIndex
var varLenIntIndex = 0;
function VarLenIntRead(s,ind) {
	varLenIntIndex = ind;
	var i = 0, sz = 1;
	for (var mb = 0; mb < 9; mb++) {
		var b = s.charCodeAt(varLenIntIndex++),
			fits = (b&128);
		i += (b&127)*sz;
		
		if (fits) break;
		sz *= 128;
	}
	return i;	
}
