
//template allows finding elements by id after id has been removed, allows deep cloning while still addressing elements by id
/**@constructor*/
function GUITemplate(id,ele) {
	this.rootElement = ele?ele:document.getElementById(id);
	this.rootElement.style.display = "none";
	
	this.cloneNodesIds = [];
	this.Build(this.rootElement);
	delete this.cloneNodesIds;
}

//recursively traverse element nodes to build template
GUITemplate.prototype.Build = function(ele) {
	if (ele.id && ele.id.length) {
		this[ele.id] = this.cloneNodesIds.slice();
		ele.removeAttribute("id");
	}

	var cn = ele.childNodes, cid = this.cloneNodesIds.length;
	this.cloneNodesIds.push(0);
	for (var i = 0; i < cn.length; i++) {
		this.cloneNodesIds[cid] = i;
		this.Build(cn[i]);
	}
	this.cloneNodesIds.length--;
}

//clone root element of template
GUITemplate.prototype.Clone = function() {
	var ele = this.rootElement.cloneNode(true);
	ele.style.removeProperty("display");
	return ele;
}

//get element from template id array, example: this.Get(rootEle, this["elementId"]) 
function GUITemplate_Get(root,cid) {
	var cn = root.childNodes;
	for (var i = 0; i < cid.length; i++) root = root.childNodes[cid[i]];
	return root;
}

//setup InputGUI for elements in root
function GUITemplate_CreateInputGUI(root,ids) {
	var eles = new Array(ids.length);
	for (var i = 0; i < ids.length; i++) eles[i] = this.Get(root,this[ids[i]]);
	return new InputGUI(ids,eles);
}
