
//Quick WebGL provides simple and easy functions for doing shader processing behind the scenes(rendering to fullscreen quad)
var gl,quickGLCanvas;



//Helper function for compiling vertex/fragment webgl shader.
function CompileShader(src,type) {
	var s = gl.createShader(type);
	gl.shaderSource(s,src);
	gl.compileShader(s);
	if (!gl.getShaderParameter(s,gl.COMPILE_STATUS)) {
		console.error("Error compiling GLSL shader: "+gl.getShaderInfoLog(s)+"\nFrom shader:");
		console.log(src);
		gl.deleteShader(s);
		return null;
	}
	return s;
}


//Rendertexture class, easy to use framebuffer object interface
var display = {framebuffer:null,width:0,height:0};

display.Bind = function() {
	gl.bindFramebuffer(gl.FRAMEBUFFER,null);
	gl.viewport(0,0,this.width,this.height);
	gl.scissor(0,0,this.width,this.height);
}


//create mesh from array of vertex positions of size dimensions(max 4), returns array buffer
function CreateMesh(dat,size) {
	var buf = gl.createBuffer();
	gl.bindBuffer(gl.ARRAY_BUFFER,buf);
	gl.bufferData(gl.ARRAY_BUFFER,buf.constructor.name==="Float32Array"?dat:new Float32Array(dat),gl.STATIC_DRAW);
	buf.meshVertexSize = size;
	buf.meshVertexCount = dat.length/size;
	return buf;
}
function UpdateMesh(buf,dat,size,dynamic) {
	gl.bindBuffer(gl.ARRAY_BUFFER,buf);
	gl.bufferData(gl.ARRAY_BUFFER,dat,dynamic?gl.DYNAMIC_DRAW:gl.STATIC_DRAW);
	buf.meshVertexSize = size;
	buf.meshVertexCount = dat.length/size;
}
function BindMesh(buf) {
	gl.bindBuffer(gl.ARRAY_BUFFER,buf);
	gl.vertexAttribPointer(0,buf.meshVertexSize,gl.FLOAT,false,buf.meshVertexSize*4,0);
}
function DeleteMesh(buf) {
	gl.deleteBuffer(buf);
}


//Initialize quick WebGL canvas and gl context, returns false if failed
var fullscreenTriangle;
function InitializeQuickWebGL(canvas) {
	quickGLCanvas = canvas;
	quickGLCanvas.width = quickGLCanvas.height = 1;
	
	//create webgl context
	var gopt = {"alpha":false, "antialias":false, "depth":false, "failIfMajorPerformanceCaveat":false, "preserveDrawingBuffer":false};
	gl = quickGLCanvas.getContext("webgl2",gopt);
	if (!gl) return false;
	
	gl.enable(gl.SCISSOR_TEST);
	gl.enable(gl.CULL_FACE);
	gl.cullFace(gl.BACK);
	gl.disable(gl.DEPTH_TEST);
	gl.depthMask(false);
	gl.clearDepth(1.0);
	gl.disable(gl.STENCIL_TEST);
	gl.disable(gl.BLEND);
	gl.enable(gl.SCISSOR_TEST);
	gl.colorMask(true,true,true,true);
	gl.enable(gl.SCISSOR_TEST);
	
	gl.enableVertexAttribArray(0);
	
	//load fullscreen triangle mesh for rendering
	fullscreenTriangle = CreateMesh([2,1, -2,1, 0,-3],2);
	
	return true;
}
